function application(S)
% ***	
% ***	 Copyright 2006, SUNY Buffalo. All rights reserved.
% ***
% ***    This program is distributed in the hope that it will be useful,
% ***    but WITHOUT ANY WARRANTY; without even the implied warranty of
% ***    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
% ***
% ***

% Written by Muzammil Hussain, mh69@cse.buffalo.edu, Initially adapted from ISIS Prowler
% code
% 


% DO NOT edit simulator code (lines that begin with S;)


S; %%%%%%%%%%%%%%%%%%%   housekeeping  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
S;      persistent app_data 
S;      global ID t
S;      [t, event, ID, data]=get_event(S);
S;      [topology, mote_IDs]=prowler('GetTopologyInfo');
S;      ix=find(mote_IDs==ID);
S;      if ~strcmp(event, 'Init_Application') 
S;         try memory=app_data{ix}; catch memory=[]; end, 
S;      end
S; %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                                          %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%         APPLICATION STARTS               %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%               HERE                       %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                                          %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv%

global fid;
global array_excited_nodes;

TRUE=1;
FALSE = 0;

%Message types
READ_REQ_MSG = 1;
READ_ANS_MSG = 2;

WRITE_ALL_REQ_MSG = 3;
WRITE_ACK_MSG = 4;

CANCEL_WRITE_MSG = 6;
CANCEL_WRITE_ACK_MSG = 7;

ACQUIRE_READ_LOCK_MSG = 8;
NOTIFY_READ_LOCK_GRANTED_MSG = 9;
ACQUIRE_WRITE_LOCK_MSG = 10;
NOTIFY_WRITE_LOCK_GRANTED_MSG = 11;

RELEASE_READ_LOCKS_MSG = 12;
RELEASE_WRITE_LOCK_MSG = 13;


MICA2_PACKET_LENGTH_BYTE = 36; % From BMAC paper
sim_params('set','MAC_PACKET_LENGTH',MICA2_PACKET_LENGTH_BYTE * 8);
sim_params('set','RADIO_SS_VAR_CONST',0);
sim_params('set','RADIO_SS_VAR_RAND',0);
sim_params('set','RECEPTION_LIMIT',0.22);   
sim_params('set','TR_ERROR_PROB',0);
sim_params('set','MAC_MIN_WAITING_TIME',200);
sim_params('set','MAC_RAND_WAITING_TIME',128);
sim_params('set','MAC_MIN_BACKOFF_TIME',(sim_params('get','MAC_PACKET_LENGTH')/2));
sim_params('set','MAC_RAND_BACKOFF_TIME',(sim_params('get','MAC_PACKET_LENGTH')/2));


WT_MAX = sim_params('get','MAC_MIN_WAITING_TIME') + sim_params('get','MAC_RAND_WAITING_TIME');
BT_MAX = sim_params('get','MAC_MIN_BACKOFF_TIME') + sim_params('get','MAC_RAND_BACKOFF_TIME');

NO_OF_RETRANSMISSIONS = 10; %>=1 $$$$  if 1 then no retransmissions

%Spacing between message requests

WACK_DELTA = NO_OF_RETRANSMISSIONS*WT_MAX + (NO_OF_RETRANSMISSIONS-1)*BT_MAX + sim_params('get','MAC_PACKET_LENGTH');
READ_REQUEST_SPACING = WACK_DELTA;
CACK_DELTA = WACK_DELTA;

WRITE_COMMIT_DELTA = WACK_DELTA + CACK_DELTA;

MAX_CANCEL_RETRIES = 10;

SECOND_IN_BIT_TIME = 40000;
PERIOD_CLEANUP_INTERVAL = 2 * SECOND_IN_BIT_TIME;

MAX_NUM_NEIGHBOURS = 8;
MAX_NUM_TRANSACTIONS = 1;

var_name_list = 'a';
var_value_list = 1:length(var_name_list);
VAR_RANGE_LENGTH = length(var_name_list);

TIMER_DELTA = MAX_NUM_NEIGHBOURS* WACK_DELTA;

LOCK_ACQ_TIME = MAX_NUM_NEIGHBOURS * WACK_DELTA; 
READ_TIME = LOCK_ACQ_TIME;

MAX_READ_LOCK_TIME = 4 * (MAX_NUM_NEIGHBOURS * WACK_DELTA);
MAX_WRITE_LOCK_TIME = 2 * (MAX_NUM_NEIGHBOURS * WACK_DELTA);

switch event
case 'Init_Application'
    signal_strength=1;
    
    %%%%%%%%%%%%%%   Memory should be initialized here  %%%%%%%%%%%%%%%%%
    memory=struct('signal_strength', signal_strength,'neigh_list',[],'multicast_read_nlist',[],'variables_to_read',[],'multicast_write_nlist',[],'variables_to_write',[],'values_to_write',[],'answering_nodes_index',[],'write_acks',[],'var_list',[],'cack_var_ref_list',[],'read_req_ref_struct',[],'cancel_acks',[],'cur_TID',-1,'read_ans_recvd',-1,'tran_count',-1,'num_tran_max',-1,'num_cancel_msg_resend',-1,'acq_wlock_acks',[],'acq_rlocks_acks',[],'read_lock_ack_struct',[],'write_lock_ack_struct',[]);
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    array_excited_nodes = sim_params('get_app','ENA');
    
    %array_excited_nodes = [12 13];

    if ID == 1
        fid = sim_params('get_app','log_file_fid1');
    end
    
    if(isempty(array_excited_nodes)) error(['Error ! No nodes are excited!!!']); end
    
    % Initialize all nodes to have variables a to z
    % Assign values from  1 to 24 for each

    memory = SetUpVariables(memory,var_name_list,var_value_list*ID);

    memory.cur_servicing_TID = -1;
    
    if(~isempty(find(array_excited_nodes == ID)))
        memory.num_tran_max = MAX_NUM_TRANSACTIONS;

        memory.tran_count = 1;
        %Initialzing the TID for the new transaction
        memory.cur_TID = str2num(strcat(num2str(ID),num2str(t)));
        disp1(sprintf('%d:%d: Starting new transaction %d with TID : %d',ID,t, memory.tran_count, memory.cur_TID)); 

        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'START');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 
        
        grid_size =  sim_params('get','GRID_SQUARE_SIZE_LENGTH');
        memory.neigh_list = GetNeighbours(ID, grid_size);
        memory = DecideReadWriteNeighbours(ID,memory);
        
        %%%%% Acquire Locks %%%%%%
        if ~isempty(memory.multicast_read_nlist)
            memory=SendAcquireReadLock(ID,t,memory,ACQUIRE_READ_LOCK_MSG,READ_REQUEST_SPACING,fid);
            %%% Send AcquireWriteLock message on receipt of all lock
            %%% granted acks            
        else
            memory=SendAcquireWriteLock(ID,t,memory,ACQUIRE_WRITE_LOCK_MSG,WACK_DELTA,fid);
        end
            
    end
    
case 'Packet_Sent'

case 'Packet_Received'
    LED('red off');
    LED('green on');
    msg=data.data;
    switch msg.type
        case ACQUIRE_READ_LOCK_MSG
            if ~(isempty(find(msg.read_nlist == ID)))
                index1 = find(msg.read_nlist == ID);
                found = FALSE;
                for i = 1:length(memory.var_list)
                    if (msg.var_to_be_read(index1) == memory.var_list(i).name)
                        found = TRUE;
                        %Chk if there is a write lock
                        if (isempty(memory.var_list(i).write_lock))
                            %Give a read lock to tid
                            memory.var_list(i).read_locks = [memory.var_list(i).read_locks msg.TID];
                            memory.var_list(i).read_locks_exp_time = [memory.var_list(i).read_locks_exp_time (t + MAX_READ_LOCK_TIME)];
                            Set_Alarm_Read_Lock_Expiry_Timer(t + MAX_READ_LOCK_TIME);
                            memory.var_list(i).read_locks;
                            
                            %Schedule a answer
                             new_i = length(memory.read_lock_ack_struct) + 1;
                             memory.read_lock_ack_struct(new_i).requesting_node = msg.ID;
                             memory.read_lock_ack_struct(new_i).requesting_tid = msg.TID;
                             memory.read_lock_ack_struct(new_i).requested_variable = msg.var_to_be_read(index1);
                             memory.read_lock_ack_struct(new_i).exp_time = t + READ_REQUEST_SPACING*(index1-1);
                
                             Set_Alarm_Send_Read_Locks_Acks_Timer(t + READ_REQUEST_SPACING*(index1-1));
                        else
                            %Ignore as read lock not possible
                            disp1(sprintf('%d:%d: Rejecting request of read lock by %d currently held by %d',ID,t,msg.TID,memory.var_list(i).write_lock));                            
                        end
                    end
                end
                
                if found == FALSE
                    error('Variable not found')
                end
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% case ACQUIRE_READ_LOCK_MSG %%%%%%
                
        case NOTIFY_READ_LOCK_GRANTED_MSG
            if msg.dest_id == ID
                  if isempty(find(memory.multicast_read_nlist == msg.ID))
                    error(['Recived read lock ack from node %d who I did not asked to write' msg.ID])
                  end

                  index = find(memory.multicast_read_nlist == msg.ID);
                  %disp1(sprintf('%d: Index in the write ack : %d',ID,index));

                  %To ensure that the written nodes answer in the order
                  %specified in the write-all message
                  for i = 1:index-1    
                      if memory.acq_rlocks_acks(i) == inf
                         % disp1(sprintf('%d: Received out of turn write ack from %d when %d has not yet replied %d',ID,msg.ID,memory.multicast_write_nlist(i)));
                      %    error(['Out of turn read answer received from : ' msg.ID])
                      end
                  end

                  %Mark that the ack has been received from the node
                  %(index)
                  memory.acq_rlocks_acks(index) = 1;
                  disp1(sprintf('%d:%d: Received read lock ack from %d',ID,t,msg.ID));

                  if index == length(memory.acq_rlocks_acks) && isempty(find(memory.acq_rlocks_acks == inf))
                    %Finished acquiring read locks 
                    %Send read request
                    memory=SendReadReq(ID,t,memory,READ_REQ_MSG,READ_REQUEST_SPACING,fid);
                  end
            end
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% case NOTIFY_READ_LOCK_GRANTED_MSG %%%%%%%
            
        case ACQUIRE_WRITE_LOCK_MSG
            if ~(isempty(find(msg.write_nlist == ID)))
                index1 = find(msg.write_nlist == ID);
                
                found = FALSE;
                for i = 1:length(memory.var_list)
                    if (msg.var_to_be_written(index1) == memory.var_list(i).name)
                        found = TRUE;
                        %Chk if there is a write lock or a read lock or a
                        %read lock for the same TID
                        %If there are no read locks and write locks on the
                        %item
                        %Or if 
                        if ((isempty(memory.var_list(i).write_lock)) && (isempty(memory.var_list(i).read_locks))) || ((isequal(memory.var_list(i).read_locks, [msg.TID])) && (length(memory.var_list(i).read_locks) == 1))
                            %Give a write lock to tid
                            memory.var_list(i).write_lock(1) = msg.TID;
                            memory.var_list(i).write_lock_exp_time = t + MAX_WRITE_LOCK_TIME;
                            Set_Alarm_Write_Lock_Expiry_Timer(t + MAX_WRITE_LOCK_TIME);
                            
                            if ~isempty(memory.var_list(i).read_locks) && isequal(memory.var_list(i).read_locks, [msg.TID]) && (length(memory.var_list(i).read_locks) == 1)
                                disp1(sprintf('%d:%d: Upgrading lock from read to write for TID %d',ID,t,msg.TID));
                            end
                            
                            %Schedule a answer
                             new_i = length(memory.write_lock_ack_struct) + 1;
                             memory.write_lock_ack_struct(new_i).requesting_node = msg.ID;
                             memory.write_lock_ack_struct(new_i).requesting_tid = msg.TID;
                             memory.write_lock_ack_struct(new_i).requested_variable = msg.var_to_be_written(index1);
                             memory.write_lock_ack_struct(new_i).exp_time = t + WACK_DELTA*(index1-1);
                
                             Set_Alarm_Send_Write_Locks_Acks_Timer(t + WACK_DELTA*(index1-1));
                        else
                            if isempty(memory.var_list(i).write_lock)
                                disp1(sprintf('%d:%d: Rejecting request of write lock by %d currently held as readers',ID,t,msg.TID));
                            else
                                disp1(sprintf('%d:%d: Rejecting request of write lock by %d currently held by %d',ID,t,msg.TID,memory.var_list(i).write_lock(1)));
                            end
                            %Ignore as write lock not possible
                        end
                    end
                end
                
                if found == FALSE
                    error('Variable not found')
                end
                
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% case ACQUIRE_WRITE_LOCK_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        case NOTIFY_WRITE_LOCK_GRANTED_MSG
            if msg.dest_id == ID %%%%1
              if isempty(find(memory.multicast_write_nlist == msg.ID))
                error(['Recived write lock ack from node %d who I did not asked to write' msg.ID])
              end

              index = find(memory.multicast_write_nlist == msg.ID);
              %disp1(sprintf('%d: Index in the write ack : %d',ID,index));

              %To ensure that the written nodes answer in the order
              %specified in the write-all message
              for i = 1:index-1    
                  if memory.acq_wlock_acks(i) == inf
                     % disp1(sprintf('%d: Received out of turn write ack from %d when %d has not yet replied %d',ID,msg.ID,memory.multicast_write_nlist(i)));
                  %    error(['Out of turn read answer received from : ' msg.ID])
                  end
              end

              %Mark that the ack has been received from the node
              %(index)
              memory.acq_wlock_acks(index) = 1;
              disp1(sprintf('%d:%d: Received write lock ack from %d',ID,t,msg.ID));

              if index == length(memory.acq_wlock_acks) && isempty(find(memory.acq_wlock_acks == inf))
                %Finished acquiring write locks 
                %Send the write request
                
                %In case the thread is only write thread then suplly own
                %variables values
                if isempty(memory.multicast_read_nlist)
                    memory.values_to_write = [];
                    for i = 1:length(memory.multicast_write_nlist)
                        for j = 1:length(memory.var_list)
                            if memory.variables_to_write(i) == memory.var_list(j).name
                                memory.values_to_write(i) = memory.var_list(j).value;
                            end
                        end
                    end
                end
                
                memory=SendWriteReq(ID,t,memory,WRITE_ALL_REQ_MSG,WACK_DELTA,fid);
              end                
            end %%%%1
              
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% case NOTIFY_WRITE_LOCK_GRANTED_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
         case RELEASE_READ_LOCKS_MSG
            if ~isempty(find(msg.read_nlist == ID))
                for i = 1:length(memory.var_list)
                    if ~isempty(find(memory.var_list(i).read_locks == msg.TID))
                        index = find(memory.var_list(i).read_locks == msg.TID);
                        disp1(sprintf('%d:%d: Releasing Read lock of  %d',ID,t,msg.TID));
                        memory.var_list(i).read_locks(index) = [];
                        memory.var_list(i).read_locks_exp_time(index) = [];
                    end            
                end
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% case RELEASE_READ_LOCKS%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

         case RELEASE_WRITE_LOCK_MSG
            if ~isempty(find(msg.write_nlist == ID))
                for i = 1:length(memory.var_list)
                    if ~isempty(memory.var_list(i).write_lock)  && (memory.var_list(i).write_lock(1) == msg.TID)
                        disp1(sprintf('%d:%d: Releasing Write lock of  %d',ID,t,msg.TID));
                        memory.var_list(i).write_lock(1) = [];
                        memory.var_list(i).write_lock_exp_time(1) = [];
                    end            
                end             
            end
            
             %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% case RELEASE_WRITE_LOCK%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         
         case READ_REQ_MSG
            % Check if the receiving node should service a read request
            if ~(isempty(find(msg.read_nlist == ID)))
                index = find(msg.read_nlist == ID);
                
                %Chk if it has a read lock
                for i=1:length(memory.var_list)
                    if (memory.var_list(i).name == msg.var_to_be_read(index)) && ~(isempty(find(memory.var_list(i).read_locks == msg.TID)))
                        new_i = length(memory.read_req_ref_struct) + 1;
                        memory.read_req_ref_struct(new_i).requesting_node = msg.ID;
                        memory.read_req_ref_struct(new_i).requesting_tid = msg.TID;
                        memory.read_req_ref_struct(new_i).requested_variable = msg.var_to_be_read(index);
                        memory.read_req_ref_struct(new_i).read_ans_exp_time = t + READ_REQUEST_SPACING*(index-1);
                        Set_Alarm_Send_Read_Request_Timer(t + READ_REQUEST_SPACING*(index-1));
                        break;
                    else
                        %Ignore as the TID does not have a read lock
                        disp1(sprintf('%d:%d: %d does not have a read lock to access',ID,t,msg.ID));
                        break;
                    end
                end
            end
                        
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case READ_REQ_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        case READ_ANS_MSG
            % If the reply is indeed from one of the requested nodes
            % for that particular transaction
            %disp1(sprintf('%d:%d: Got Read answer!!!',ID,t));
            prev_read_answers_not_recvd = FALSE;
            
            if msg.dest_id == ID && ~(isempty(find(memory.multicast_read_nlist == msg.ID))) %&& msg.TID == memory.cur_TID
                index = find(memory.multicast_read_nlist == msg.ID);
                %disp1(sprintf('%d:%d: Received answer whose Index in the read request : %d',ID,t,index));
                % Check if it has come in the right order as specified in
                % the read request
                for i = 1:index-1    
                    if memory.answering_nodes_index(i).flag == inf
                        prev_read_answers_not_recvd = TRUE;
                    end
                end

                if memory.answering_nodes_index(index).flag ~= inf
                    error(['%d trying to overwrite' msg.ID])
                end

                % Record the answer
                disp1(sprintf('%d:%d Recording answer from node %d',ID,t,msg.ID));
                memory.answering_nodes_index(index).name = msg.variable_names;
                memory.answering_nodes_index(index).value = msg.variable_values;
                memory.answering_nodes_index(index).flag = 1;

                %In case all requested nodes have replied send the
                %writeall
                if index == length(memory.multicast_read_nlist) && (prev_read_answers_not_recvd == FALSE)
                    memory.read_ans_recvd = 1;
                    PrintMessage('ACQ WL');
                    
                    read_answers = [];
                    for i = 1:length(memory.answering_nodes_index)
                        read_answers(i)= memory.answering_nodes_index(i).value;
                    end
                    
                    result=ApplyTranformation(read_answers,'PRODUCT');
                    %Set each variable in the neighbor list to be
                    %written the same value
                    memory.values_to_write = [];
                    for i=1:length(memory.multicast_write_nlist)
                        memory.values_to_write(i) = result;
                    end

                    %Start acquiring write locks
                    memory = SendAcquireWriteLock(ID,t,memory,ACQUIRE_WRITE_LOCK_MSG,WACK_DELTA,fid);
                end
            end

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case READ_ANS_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        case WRITE_ALL_REQ_MSG
            if ~(isempty(find(msg.write_nlist == ID))) 
                index = find(msg.write_nlist == ID);
          %      disp1(sprintf('%d: Servicing write request for transaction %d from node %d for variable %c',ID,msg.TID,msg.ID,msg.variables_to_write(index)));
%                 memory.updating_node = msg.ID;
%                 memory.cur_servicing_TID = msg.TID;

                for i=1:length(memory.var_list)
                    if (memory.var_list(i).name == msg.variables_to_write(index)) && ~(isempty(find(memory.var_list(i).write_lock == msg.TID)))
                        memory.var_list(i).new_value = [memory.var_list(i).new_value msg.values_to_write(index)];
                        disp1(sprintf('%d:%d: Updating variable %c to new value %d supplied by %d for transaction %d',ID,t,memory.var_list(i).name,memory.var_list(i).new_value(end),msg.ID,msg.TID));

                        %Write to the designated variable
                        Set_Write_Commit_Timer(t + length(msg.write_nlist)*WRITE_COMMIT_DELTA);
                        memory.var_list(i).commit_time = [memory.var_list(i).commit_time (t + length(msg.write_nlist)*WRITE_COMMIT_DELTA)];
                        memory.var_list(i).updating_tid = [memory.var_list(i).updating_tid msg.TID];
                        memory.var_list(i).updating_node = [memory.var_list(i).updating_node msg.ID];

                        %Send the ACK in the order of position in the write list
                        index = find(msg.write_nlist == ID);    
                        Set_Alarm_Send_Write_All_Ack_Timer(t + WACK_DELTA*(index-1));   
                        memory.var_list(i).w_ack_send_time = [memory.var_list(i).w_ack_send_time t + WACK_DELTA*(index-1)];
                        break;
                    end
                end
                % Even the servcing node also snoops at its own service
                %disp1(sprintf('%d: I am also snooping he he he !!!'));
            end

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case WRITE_ALL_REQ_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        case WRITE_ACK_MSG
            if msg.dest_id == ID
              if isempty(find(memory.multicast_write_nlist == msg.ID))
                error(['Recived ack from node %d who I did not asked to write' msg.ID])
              end

              index = find(memory.multicast_write_nlist == msg.ID);
              %disp1(sprintf('%d: Index in the write ack : %d',ID,index));

              %To ensure that the written nodes answer in the order
              %specified in the write-all message
              for i = 1:index-1    
                  if memory.write_acks(i) == inf
                     % disp1(sprintf('%d: Received out of turn write ack from %d when %d has not yet replied %d',ID,msg.ID,memory.multicast_write_nlist(i)));
                  %    error(['Out of turn read answer received from : ' msg.ID])
                  end
              end

              %Mark that the ack has been received from the node
              %(index)
              memory.write_acks(index) = 1;
              disp1(sprintf('%d:%d: Received write ack from %d',ID,t,msg.ID));

              if index == length(memory.write_acks) && isempty(find(memory.write_acks == inf))
                PrintMessage('END'); 
                disp1(sprintf('%d:%d: Received acks from all',ID,t));
                %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
                fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'END');
                %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 
                num_finished_nodes = sim_params('get_app','num_finished_nodes');
                num_finished_nodes = num_finished_nodes + 1;
                sim_params('set_app','num_finished_nodes',num_finished_nodes);
              end
            end

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case WRITE_ACK_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        case CANCEL_WRITE_MSG
            %Chk if the msg is intended for you
            if ~isempty(find(msg.multicast_nlist == ID))
                %Scan through the commit times 
                for i= 1:length(memory.var_list)
                    %Check if that particular variable is involved in the
                    %cancelled transaction
                    index1 = find(memory.var_list(i).updating_node == msg.ID);
                    if ~isempty(index1)
                        % Discard the new value and reset everything else
                        disp1(sprintf('%d:%d: Discarding new value for %c as commanded by %d',ID,t,memory.var_list(i).name,msg.ID));
                                                
                        memory.var_list(i).new_value(index1) = inf;
                        memory.var_list(i).updating_tid(index1) = -1;
                    end
                end

                index2 = find(msg.multicast_nlist == ID);
                %disp1(sprintf('%d:%d: Index in the write cancel request : %d',ID,t,index));
                Set_Alarm_Send_Cancel_Ack_Timer(t + CACK_DELTA*(index2-1));  

                %Add information to the cack_var_ref_list
                new_i = length(memory.cack_var_ref_list) + 1;
                memory.cack_var_ref_list(new_i).requesting_node = msg.ID;
                memory.cack_var_ref_list(new_i).c_ack_send_time = t + CACK_DELTA*(index2-1);
            end

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case CANCEL_WRITE_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        case CANCEL_WRITE_ACK_MSG
            if msg.dest_id == ID
              if isempty(find(memory.multicast_write_nlist == msg.ID))
                error(['Recived cancel ack from node %d who I did not asked to write' msg.ID])
              end

              index = find(memory.multicast_write_nlist == msg.ID);
           %   disp1(sprintf('%d:%d: Index in the cancel ack : %d',ID,t,index));

              %To ensure that the written nodes answer in the order
              %specified in the write-all message
              for i = 1:index-1    
                  if memory.cancel_acks(i) == inf
                      disp1(sprintf('%d:%d: Received out of turn cancel ack from %d when %d has not yet replied %d',ID,t,msg.ID,memory.multicast_write_nlist(i)));
                     % error(['Out of turn read answer received from : ' msg.ID])
                  end
              end

              %Mark that the ack has been received from the node
              %(index)
              memory.cancel_acks(index) = 1;
              disp1(sprintf('%d:%d: Received cancel ack from %d',ID,t,msg.ID));
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case CANCEL_WRITE_ACK_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        otherwise
            error(['Bad msg type received: ' msg.type])
    end
case 'Collided_Packet_Received'
    disp1(sprintf('%d:%d: Received collided packet !!',ID,t));
    LED('red on');
    LED('green off')
    
case 'Send_Read_Answer_Timer_Expired'
    % Send the values of the variables as ordered in memory.read_requested_variables
    disp1(sprintf('%d:%d: Read Answer Timer Expired !!!',ID,t));
    
    index = -1;
    for i = 1:length(memory.read_req_ref_struct)
        if memory.read_req_ref_struct(i).read_ans_exp_time == t
            index = i;
            break;
        end
    end
    
    found = FALSE;
    for i=1:length(memory.var_list)
        if memory.var_list(i).name == memory.read_req_ref_struct(index).requested_variable
            disp1(sprintf('%d:%d: Reading variable %c which gives value %d',ID,t,memory.var_list(i).name,memory.var_list(i).value));
            PrintMessage('R ANS');
            Send_Packet(radiostream(struct('ID',ID, 'type',READ_ANS_MSG,'dest_id',memory.read_req_ref_struct(index).requesting_node,'TID',memory.read_req_ref_struct(index).requesting_tid,'variable_names',memory.var_list(i).name,'variable_values',memory.var_list(i).value),memory.signal_strength));
            disp1(sprintf('%d:%d: Sending answer to %d for transaction %d',ID,t,memory.read_req_ref_struct(index).requesting_node,memory.read_req_ref_struct(index).requesting_tid));
            found = TRUE;
            %Remove the head
            memory.read_req_ref_struct(index) = [];
            break;
        end
    end
        
    if found == FALSE
        error(['Error ! Read request variable does not exist in node'])
    end

    
case 'Receive_Read_Answers_Timer_Expired'
    if memory.read_ans_recvd == 0
        disp1(sprintf('%d:%d: Read Request timing out !!!',ID,t));
        disp1(sprintf('%d:%d: Canceling transaction %d !!!',ID,t,memory.cur_TID));
        
        PrintMessage('CNCL R');

        %Release all the read locks
        SendReleaseReadLocks(ID,t,memory,RELEASE_READ_LOCKS_MSG);

        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        %fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'TIMEOUT-CANCEL');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%        

        memory.cur_TID = str2num(strcat(num2str(ID),num2str(t)));
        disp1(sprintf('%d:%d: Starting %d new transaction with TID : %d',ID,t, memory.tran_count, memory.cur_TID));
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'START');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 

        %Acquire new read locks for the transaction
        memory=SendAcquireReadLock(ID,t,memory,ACQUIRE_READ_LOCK_MSG,READ_REQUEST_SPACING,fid);
    end
%     else
         disp1(sprintf('%d:%d: Ignoring Read Answer Timer',ID,t));
%     end
        
    
case 'Write_Commit_Timer_Expired'
    %  disp1(sprintf('%d: Write Commit Timer Expired at %d',ID,t));
    % Commit the the values of variable written into the sandbox
    for i = 1:length(memory.var_list)
        % Check if the variable is to be commited at time t and commit
        % in case it should be commited at time t 
        if memory.var_list(i).commit_time(1) <= t
           if memory.var_list(i).new_value(1) ~= inf
               %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
               %fprintf(fid, '%d:%d:%s\n',memory.var_list(i).updating_tid(1),t,'COMMIT');
               %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
               disp1(sprintf('%d:%d: Commiting %d in variable %c at time %d for transaction %d',ID,t,memory.var_list(i).new_value(1),memory.var_list(i).name,t,memory.var_list(i).updating_tid(1)));
               memory.var_list(i).value = memory.var_list(i).new_value(1);
           else
               disp1(sprintf('%d:%d: Ignoring the write commit timer as the write has been cancelled',ID,t));
           end

           %Remove the  head entry
           memory.var_list(i).new_value(1) = [];
           memory.var_list(i).updating_tid(1) = [];
           memory.var_list(i).commit_time(1) = [];
           memory.var_list(i).updating_node(1) = [];
           memory.var_list(i).w_ack_send_time(1) = [];
        else
            disp1(sprintf('%d:%d: %c commit time %d current time %d',ID,t, memory.var_list(i).name,memory.var_list(i).commit_time(1), t));
        end
    end

case 'Receive_Write_All_Ack_Timer_Expired'
    %Check if it has received acks from all the members in the
    %multicast_write_nlist (by searching for a ack not yet come)
    if ~(isempty(find(memory.write_acks == inf))) 
     %   disp1(sprintf('%d:%d: Receive write ack timer expired !!!',ID,t));

        %Release all read and write locks before aborting the write
        SendReleaseReadLocks(ID,t,memory,RELEASE_READ_LOCKS_MSG);
        SendReleaseWriteLock(ID,t,memory,RELEASE_WRITE_LOCK_MSG);
     
        %Cancel the entire transaction as at least one node has not acked
        %the write_all message
        disp1(sprintf('%d:%d: Write Ack Timer Expired !!! Canceling transaction %d !',ID,t,memory.cur_TID));
        memory.num_cancel_msg_resend = memory.num_cancel_msg_resend + 1;
        PrintMessage('CNCL');
        Send_Packet(radiostream(struct('ID',ID, 'type',CANCEL_WRITE_MSG,'aborted_tid',memory.cur_TID,'multicast_nlist',memory.multicast_write_nlist),memory.signal_strength));        

        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        %fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'TIMEOUT-CANCEL');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%

        memory.cancel_acks = [];
        for i=1:length(memory.multicast_write_nlist)
            memory.cancel_acks = [memory.cancel_acks inf];
        end

        Set_Alarm_Receive_Cancel_Ack_Timer(t + length(memory.multicast_write_nlist)*CACK_DELTA);
    else
        % Release all read and write locks after a transaction
        SendReleaseReadLocks(ID,t,memory,RELEASE_READ_LOCKS_MSG);
        SendReleaseWriteLock(ID,t,memory,RELEASE_WRITE_LOCK_MSG);
        
        disp1(sprintf('%d: Transaction %d completes at time %d',ID,memory.cur_TID,t));
        %start new transaction thread 
        if memory.tran_count < memory.num_tran_max
            memory.tran_count = memory.tran_count + 1;

            %Initialzing the TID for the new transaction
            memory.cur_TID = str2num(strcat(num2str(ID),num2str(t)));
            disp1(sprintf('%d:%d: Starting %d new transaction with TID : %d',ID,t, memory.tran_count, memory.cur_TID));
            %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
            fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'START');
            %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 
            memory = DecideReadWriteNeighbours(ID,memory,VAR_RANGE_LENGTH);
            %In case it is a read-write thread structure
            if ~isempty(memory.multicast_read_nlist)
                %Send the read request
                memory=SendAcquireReadLock(ID,t,memory,ACQUIRE_READ_LOCK_MSG,READ_REQUEST_SPACING,fid);
            else
                %Write values of the own same name variables to neighbours
                values_to_write = [];
                for i = 1:length(memory.multicast_write_nlist)
                    for j = 1:length(memory.var_list)
                        if memory.variables_to_write(i) == memory.var_list(j).name
                            memory.values_to_write(i) = memory.var_list(j).value;
                        end
                    end
                end                
                %Send the write request
                memory = SendAcquireWriteLock(ID,t,memory,ACQUIRE_WRITE_LOCK_MSG,WACK_DELTA,fid);
            end
        end        
    end
    %Else ignore the timer
    
case 'Send_Write_All_Ack_Timer_Expired'
    disp1(sprintf('%d:%d: Send write ack timer expired !!!',ID,t));
    for i = 1:length(memory.var_list)
        index = find(memory.var_list(i).w_ack_send_time == t);

        if ~isempty(index)
            PrintMessage('W ACK');
            disp1(sprintf('%d:%d: Sending write ack to %d',ID,t,memory.var_list(i).updating_node(index)));
            Send_Packet(radiostream(struct('ID',ID, 'type',WRITE_ACK_MSG,'dest_id',memory.var_list(i).updating_node(index)),memory.signal_strength));    
            break;
        end
    end
    
case 'Send_Cancel_Ack_Timer_Expired'
    for i = 1:length(memory.cack_var_ref_list)
        if(memory.cack_var_ref_list(i).c_ack_send_time == t)
            PrintMessage('CACK');
            disp1(sprintf('%d:%d: Sending cancel ack to %d',ID,t,memory.cack_var_ref_list(i).requesting_node));
            Send_Packet(radiostream(struct('ID',ID, 'type',CANCEL_WRITE_ACK_MSG,'dest_id',memory.cack_var_ref_list(i).requesting_node),memory.signal_strength));             
            break;
        end
    end
    memory.cack_var_ref_list(i) = [];
    
    
case 'Receive_Cancel_Ack_Timer_Expired'
    %Check if it has received cancel acks from all the members in the
    %multicast_write_nlist (by searching for a cancel ack not yet come)
    if memory.num_cancel_msg_resend < MAX_CANCEL_RETRIES && ~(isempty(find(memory.cancel_acks == inf)))
        disp1(sprintf('%d:%d: Receive cancel ack timer expired !!!',ID,t));
        %Resend the cancel message again when at least one node has not acked
        %the cancel message
        memory.num_cancel_msg_resend = memory.num_cancel_msg_resend + 1;
        disp1(sprintf('%d:%d: Resending the cancel message !!!',ID,t));
        PrintMessage('CNCL');
        Send_Packet(radiostream(struct('ID',ID, 'type',CANCEL_WRITE_MSG,'aborted_tid',memory.cur_TID,'multicast_nlist',memory.multicast_write_nlist),memory.signal_strength));        
        Set_Alarm_Receive_Cancel_Ack_Timer(t + length(memory.multicast_write_nlist)*CACK_DELTA);
        
        memory.cancel_acks = [];
        for i=1:length(memory.multicast_write_nlist)
            memory.cancel_acks = [memory.cancel_acks inf];
        end        
    else
        %start new transaction thread 
%         if memory.tran_count < memory.num_tran_max
%             memory.tran_count = memory.tran_count + 1;

        %Initialzing the TID for the new transaction
        memory.cur_TID = str2num(strcat(num2str(ID),num2str(t)));
        disp1(sprintf('%d:%d: Starting new transaction %d with TID : %d to do the same',ID,t, memory.tran_count, memory.cur_TID));
        %memory = DecideReadWriteNeighbours(ID,memory,VAR_RANGE_LENGTH);
        
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'START');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 
        
        %In case it is a read-write thread structure
        if ~isempty(memory.multicast_read_nlist)
            %Send the read request
            memory=SendReadReq(ID,t,memory,READ_REQ_MSG,READ_REQUEST_SPACING,fid);
        else
            %Send the write request
            %Write values of the own same name variables to neighbours
            memory.values_to_write = [];
            for i = 1:length(memory.multicast_write_nlist)
                for j = 1:length(memory.var_list)
                    if memory.variables_to_write(i) == memory.var_list(j).name
                        memory.values_to_write(i) = memory.var_list(j).value;
                    end
                end
            end

            %Send the write request
            memory=SendWriteReq(ID,t,memory,WRITE_ALL_REQ_MSG,WACK_DELTA,fid);
        end
  %  end
    end
    %Else ignore the timer

case 'Receive_Read_Lock_Ack_Timer_Expired'

    if (~isempty(find(memory.acq_rlocks_acks == inf)))
        %Release all the read locks
        SendReleaseReadLocks(ID,t,memory,RELEASE_READ_LOCKS_MSG);
        
        %Restart the transaction
        %Initialzing the TID for the new transaction
        memory.cur_TID = str2num(strcat(num2str(ID),num2str(t)));
        disp1(sprintf('%d:%d: Starting new transaction %d with TID : %d to do the same',ID,t, memory.tran_count, memory.cur_TID));

        %In case it is a read-write thread structure
        if ~isempty(memory.multicast_read_nlist)
            %Send the acquire read lock request
            memory=SendAcquireReadLock(ID,t,memory,ACQUIRE_READ_LOCK_MSG,READ_REQUEST_SPACING,fid);
        else
            %Send the acqure write lock request
            memory=SendAcquireWriteLock(ID,t,memory,ACQUIRE_WRITE_LOCK_MSG,WACK_DELTA,fid);
        end
    else
        disp1(sprintf('%d:%d: Ignoring Receive_Read_Lock_Ack_Timer_Expired as all locks received',ID,t));
    end

case 'Receive_Write_Lock_Ack_Timer_Expired'
    if (~isempty(find(memory.acq_wlock_acks == inf)))
        %Release all the read and write locks
        SendReleaseReadLocks(ID,t,memory,RELEASE_READ_LOCKS_MSG);
        SendReleaseWriteLock(ID,t,memory,RELEASE_WRITE_LOCK_MSG);
        
        %Start the congestion control timer of 0,1,2
        r= 0;
        disp1(sprintf('%d:%d: Postponing transaction by %d times',ID,t,r));
        rtime = READ_REQUEST_SPACING * (length(memory.multicast_read_nlist) + 2);
        wctime = WRITE_COMMIT_DELTA * (length(memory.multicast_write_nlist) + 2);
        max_transaction_time = rtime + wctime;
        Set_Congestion_Transaction_Backoff_Timer(t + (r*max_transaction_time));  
    else
        disp1(sprintf('%d:%d: Ignoring Receive_Write_Lock_Ack_Timer_Expired as all locks received',ID,t));
    end
    
case 'Congestion_Transaction_Backoff_Timer_Expired'
        %start new transaction thread 
        disp1(sprintf('%d:%d: Starting back after backoff',ID,t));
    
        %Restart the transaction
        %Initialzing the TID for the new transaction
        memory.cur_TID = str2num(strcat(num2str(ID),num2str(t)));
        disp1(sprintf('%d:%d: Starting new transaction %d with TID : %d to do the same',ID,t, memory.tran_count, memory.cur_TID));
        %memory = DecideReadWriteNeighbours(ID,memory,VAR_RANGE_LENGTH);

        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'START');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 

        %In case it is a read-write thread structure
        if ~isempty(memory.multicast_read_nlist)
            %Send the acquire read lock request
            memory=SendAcquireReadLock(ID,t,memory,ACQUIRE_READ_LOCK_MSG,READ_REQUEST_SPACING,fid);
        else
            %Send the acqure write lock request
            memory=SendAcquireWriteLock(ID,t,memory,ACQUIRE_WRITE_LOCK_MSG,WACK_DELTA,fid);
        end

    
case 'Send_Read_Lock_Ack_Timer_Expired'
    for i = 1:length(memory.read_lock_ack_struct)
        if(memory.read_lock_ack_struct(i).exp_time == t)
            PrintMessage('GIVE RL');
            disp1(sprintf('%d:%d: Sending read lock granted to %d',ID,t,memory.read_lock_ack_struct(i).requesting_node));
            Send_Packet(radiostream(struct('ID',ID, 'type',NOTIFY_READ_LOCK_GRANTED_MSG,'dest_id',memory.read_lock_ack_struct(i).requesting_node),memory.signal_strength));             
            break;
        end
    end
    memory.read_lock_ack_struct(i) = [];
    
case 'Send_Write_Lock_Ack_Timer_Expired'
    for i = 1:length(memory.write_lock_ack_struct)
        if(memory.write_lock_ack_struct(i).exp_time == t)
            PrintMessage('GIVE WL');
            disp1(sprintf('%d:%d: Sending write lock granted to %d',ID,t,memory.write_lock_ack_struct(i).requesting_node));
            Send_Packet(radiostream(struct('ID',ID,'type',NOTIFY_WRITE_LOCK_GRANTED_MSG,'dest_id',memory.write_lock_ack_struct(i).requesting_node),memory.signal_strength));             
            break;
        end
    end
    memory.write_lock_ack_struct(i) = [];
    
case 'Release_Read_Lock_Timer_Expired'
    for i = 1:length(memory.var_list)
        if ~isempty(memory.var_list(i).read_locks_exp_time) && memory.var_list(i).read_locks_exp_time(1) == t
            disp1(sprintf('%d:%d: A read lock has expired after MAX waiting',ID,t));
            memory.var_list(i).read_locks(1) = [];
            memory.var_list(i).read_locks_exp_time(1) = [];
        end            
    end
    
case 'Release_Write_Lock_Timer_Expired'
    for i = 1:length(memory.var_list)
        if ~isempty(memory.var_list(i).write_lock_exp_time) && (memory.var_list(i).write_lock_exp_time(1) == t)
            disp1(sprintf('%d:%d: A write lock has expired after MAX waiting',ID,t));
            memory.var_list(i).write_lock(1) = [];
            memory.var_list(i).write_lock_exp_time(1) = [];
        end            
    end
    
case 'GuiInfoRequest'
    if ~isempty(memory)
        disp1(sprintf('Memory Dump of mote ID# %d:\n',ID)); 
    else
        disp1(sprintf('No memory dump available for node %d.\n',ID)); 
    end
    
case 'Application_Stopped'
   % this event is called when simulation is stopped/suspended 
   
case 'Application_Finished'
   % this event is called when simulation is finished 
    
otherwise
    error(['Bad event name for application: ' event])
end

%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                                          %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%         APPLICATION ENDS                 %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%               HERE                       %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                                          %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
S; %%%%%%%%%%%%%%%%%%%%%% housekeeping %%%%%%%%%%%%%%%%%%%%%%%%%%%
S;        app_data{ix}=memory;
S; %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                           %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%        COMMANDS           %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                           %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function b=Send_Packet(data);
global ID t
radio=prowler('GetRadioName');
b=feval(radio, 'Send_Packet', ID, data, t);

function b=Set_Alarm_Send_Read_Request_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Send_Read_Answer_Timer_Expired', ID));

function b=Set_Alarm_Receive_Read_Answers_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Receive_Read_Answers_Timer_Expired', ID));

function b=Set_Receive_Write_All_Ack_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Receive_Write_All_Ack_Timer_Expired', ID));

function b=Set_Alarm_Send_Write_All_Ack_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Send_Write_All_Ack_Timer_Expired', ID));

function b=Set_Alarm_Send_Cancel_Ack_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Send_Cancel_Ack_Timer_Expired', ID));

function b=Set_Alarm_Receive_Cancel_Ack_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Receive_Cancel_Ack_Timer_Expired', ID));

function b=Set_Write_Commit_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Write_Commit_Timer_Expired', ID));

function b=Set_Misc_Alarm_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Misc_Alarm_Expired', ID));

function b=Set_Snoop_Commit_Alarm_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Snoop_Commit_Timer_Expired', ID));

function b=Set_Snoop_Extended_Commit_Alarm_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Snoop_Extended_Commit_Timer_Expired', ID));

function b=Set_Periodic_Cleanup_Alarm_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Periodic_Cleanup_Timer_Expired', ID));

function b=Set_Alarm_Receive_Read_Locks_Acks_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Receive_Read_Lock_Ack_Timer_Expired', ID));

function b=Set_Alarm_Receive_Write_Locks_Acks_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Receive_Write_Lock_Ack_Timer_Expired', ID));

function b=Set_Alarm_Send_Read_Locks_Acks_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Send_Read_Lock_Ack_Timer_Expired', ID));

function b=Set_Alarm_Send_Write_Locks_Acks_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Send_Write_Lock_Ack_Timer_Expired', ID));

function b=Set_Alarm_Read_Lock_Expiry_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Release_Read_Lock_Timer_Expired', ID));

function b=Set_Alarm_Write_Lock_Expiry_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Release_Write_Lock_Timer_Expired', ID));

function b=Set_Congestion_Transaction_Backoff_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Congestion_Transaction_Backoff_Timer_Expired', ID));

function PrintMessage(msg)
global ID
prowler('TextMessage', ID, msg)

function LED(msg)
global ID
prowler('LED', ID, msg)

function DrawLine(command, varargin)
switch lower(command)
case 'line'
    prowler('DrawLine', varargin{:})
case 'arrow'
    prowler('DrawArrow', varargin{:})
case 'delete'
    prowler('DrawDelete', varargin{:})
otherwise
    error('Bad command for DrawLine.')
end

function array =GetRandomArray(n,max)
    count = 1;
    array = [];
    while(count<n+1)
      x = ceil(max*rand);
      if isempty(find(array == x))
        array(count) = x;
        count = count + 1;
      end
    end
    
%x is the number of randomly chosen nodes needed
%N is the side of the N*N grid
%For example (5,5) would return an array of 5 randomly chosen nodes from a
%square grid of 25 nodes
function array = GetNeighbours(x,N)
    array = [];
    left = x - N;
    right = x + N;
    % Add the left neighbour if possible
    if(left >= 1)
        array = [array left];
    end
    % Add the right neighbour if possible
    if(right <= N*N)
        array = [array right];
    end
    
    %Add the northern/sourthern neighbour
    if( mod(x, N) == 0)
        array = [array x-1]; % Add southern neighbour if x is topmost
    elseif(mod(x, N) == 1)
        array = [array x+1];
    else
        array = [array x-1 x+1];
    end
    
    if(left >= 1)
        %Add the northern/sourthern neighbour of left
        if( mod(left, N) == 0)
            array = [array left-1]; % Add southern neighbour if x is topmost
        elseif(mod(left,N) == 1)
            array = [array left+1];
        else
            array = [array left-1 left+1];
        end
    end 
    
    if(right <= N*N)
        %Add the northern/sourthern neighbour of right
        if( mod(right,N) == 0)
            array = [array right-1]; % Add southern neighbour if x is topmost
        elseif(mod(right,N) == 1)
            array = [array right+1];
        else
            array = [array right-1 right+1];
        end
    end     
    array = sort(array);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function mem=SetUpVariables(mem,name_list,value_list)
    len = length(name_list);
    for i = 1 : len
        mem.var_list(i).name = name_list(i);
        mem.var_list(i).value = value_list(i);
        mem.var_list(i).new_value = [];
        mem.var_list(i).commit_time = [];
        mem.var_list(i).w_ack_send_time = [];
        mem.var_list(i).updating_tid = [];
        mem.var_list(i).updating_node = [];
        mem.var_list(i).read_locks = [];
        mem.var_list(i).read_locks_exp_time = [];
        mem.var_list(i).write_lock = [];
        mem.var_list(i).write_lock_exp_time = [];
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function memory = DecideReadWriteNeighbours(ID,memory)

    %Decide the read neighbours
    memory = DecideReadNeighbours(ID,memory);
    while(length(memory.multicast_read_nlist) == 0)
        memory = DecideReadNeighbours(ID,memory);
    end

    %Decide the write neighbours % At least one
    memory = DecideWriteNeighbours(ID,memory);
    while(length(memory.multicast_write_nlist) == 0)
        memory = DecideWriteNeighbours(ID,memory);
    end
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function memory = DecideReadNeighbours(ID,memory)
    read_trans_matrix = zeros(1,length(memory.neigh_list));
    
    %Decide the read neighbours
    for i=1:length(read_trans_matrix)
        %Randomly choose the read neigbours with 1 being they are in the
        %list
        read_trans_matrix(i) = round(rand);
        if (read_trans_matrix(i) == 1)
            memory.variables_to_read = [memory.variables_to_read 'a'];
        end
    end
    memory.multicast_read_nlist = memory.neigh_list(find(read_trans_matrix == 1));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function memory = DecideNReadNeighbours(ID,memory,n)
    count = 0;
    memory.multicast_read_nlist = [];
    memory.variables_to_read = [];
    
    while (count < n)
        x = neigh_list(ceil(rand*length(neigh_list)));
        var_x = sprintf('%c',96 + ceil(rand * VAR_RANGE_LENGTH));
        if isempty(find(memory.multicast_read_nlist == x))
            memory.multicast_read_nlist = [memory.multicast_read_nlist x];
            memory.variables_to_read = [memory.variables_to_read var_x];
            count = count + 1;
        end
    end
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function memory = DecideWriteNeighbours(ID,memory)
    write_trans_matrix = zeros(1,length(memory.multicast_read_nlist));
    
    %Decide the write neighbours
    for i=1:length(write_trans_matrix)
        %If there is only one read neighbour then write to that without
        %choosing
        if length(memory.multicast_read_nlist) == 1
            write_trans_matrix(1) = 1;
            memory.variables_to_write = [memory.variables_to_write 'a'];
            break;
        else
            %Randomly choose the write neigbours with 1 being they are in the
            %list
            write_trans_matrix(i) = round(rand);        
            if (write_trans_matrix(i) == 1)
                memory.variables_to_write = [memory.variables_to_write 'a'];
            end
        end
    end
    memory.multicast_write_nlist = memory.multicast_read_nlist(find(write_trans_matrix == 1));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function memory = DecideNWriteNeighbours(ID,memory,n)
    count = 0;
    memory.multicast_write_nlist = [];
    memory.variables_to_write = [];
    
    while (count < n)
        x = neigh_list(ceil(rand*length(neigh_list)));
        var_x = sprintf('%c',96 + ceil(rand * VAR_RANGE_LENGTH));
        if isempty(find(memory.multicast_write_nlist == x))
            memory.multicast_write_nlist = [memory.multicast_write_nlist x];
            memory.variables_to_write = [memory.variables_to_write var_x];
            count = count + 1;
        end
    end
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function result=ApplyTranformation(list,oper_type)
    switch oper_type
        case 'PRODUCT'
            result = prod(list);
        case 'SUM'
            result = sum(list);
        case 'MAX'
            result = max(list);
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
function memory = SendReadReq(ID,t,memory,READ_REQ_MSG,READ_REQUEST_SPACING,fid)
    disp1(sprintf('%d: Sending read request',ID));
    PrintMessage('R REQ');
    printIntArray(memory.multicast_read_nlist);
    
    Send_Packet(radiostream(struct('ID',ID, 'type',READ_REQ_MSG,'TID',memory.cur_TID,'read_nlist',memory.multicast_read_nlist,'var_to_be_read',memory.variables_to_read),memory.signal_strength));

    memory.read_ans_recvd = 0;
    %Start a timer for the sender to wait for all the ACKS
    Set_Alarm_Receive_Read_Answers_Timer(t+ length(memory.multicast_read_nlist)*READ_REQUEST_SPACING);

    %Initialize the answer array
    memory.answering_nodes_index = [];
    for i=1:length(memory.multicast_read_nlist)
        memory.answering_nodes_index(i).flag = inf;
        memory.answering_nodes_index(i).name = '';
        memory.answering_nodes_index(i).value = inf;
    end
    
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    for i=1:length(memory.multicast_read_nlist)
        %fprintf(fid, '%d:%d:%s:%d:%c\n', memory.cur_TID,t,'READ',memory.multicast_read_nlist(i),memory.variables_to_read(i));
    end
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
function memory = SendWriteReq(ID,t,memory,WRITE_ALL_REQ_MSG,WACK_DELTA,fid)
    PrintMessage('W ALL');
    disp1(sprintf('%d: Sending write req',ID));
    printIntArray(memory.multicast_write_nlist);
    
    %Initialize the write ack receive array
    memory.write_acks = [];
    for i=1:length(memory.multicast_write_nlist)
        memory.write_acks = [memory.write_acks inf];
    end

   memory.num_cancel_msg_resend = 0;
   
   %Sending the write-all message to subset of neighbors
   Send_Packet(radiostream(struct('ID',ID, 'type',WRITE_ALL_REQ_MSG,'TID',memory.cur_TID,'write_nlist',memory.multicast_write_nlist,'variables_to_write',memory.variables_to_write,'values_to_write',memory.values_to_write),memory.signal_strength));

   %Start a timer for the sender to wait for all the ACKS
   Set_Receive_Write_All_Ack_Timer(t+ length(memory.multicast_write_nlist)*WACK_DELTA);

   %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    write_string = [];
    for i=1:length(memory.multicast_write_nlist)
        write_string = strcat(write_string,sprintf('::%d:%c',memory.multicast_write_nlist(i),memory.variables_to_write(i)));
    end
    %fprintf(fid, '%d:%d:%s:%s\n', memory.cur_TID,t,'WRITE',write_string);
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   

function  memory=SendAcquireReadLock(ID,t,memory,ACQUIRE_READ_LOCK_MSG,READ_REQUEST_SPACING,fid)
    PrintMessage('ACQ R');
    disp1(sprintf('%d: Sending acquire read req',ID));
    printIntArray(memory.multicast_read_nlist);

    Send_Packet(radiostream(struct('ID',ID, 'type',ACQUIRE_READ_LOCK_MSG,'TID',memory.cur_TID,'read_nlist',memory.multicast_read_nlist,'var_to_be_read',memory.variables_to_read),memory.signal_strength));

    %Initialize read lock array.
    memory.acq_rlocks_acks = [];
    for i=1:length(memory.multicast_read_nlist)
        memory.acq_rlocks_acks = [memory.acq_rlocks_acks inf];
    end    
    
    memory.read_ans_recvd = 0;
    %Start a timer for the sender to wait for all the ACKS
    Set_Alarm_Receive_Read_Locks_Acks_Timer(t+ (length(memory.multicast_read_nlist)+ 1)*READ_REQUEST_SPACING);
    
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    for i=1:length(memory.multicast_read_nlist)
        %fprintf(fid, '%d:%d:%s:%d:%c\n', memory.cur_TID,t,'ACQ RLOCK',memory.multicast_read_nlist(i),memory.variables_to_read(i));
    end
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    
function  memory=SendAcquireWriteLock(ID,t,memory,ACQUIRE_WRITE_LOCK_MSG,WACK_DELTA,fid)
    PrintMessage('ACQ W');
    disp1(sprintf('%d: Sending Acquire write req',ID));
    printIntArray(memory.multicast_write_nlist);
    
    Send_Packet(radiostream(struct('ID',ID, 'type',ACQUIRE_WRITE_LOCK_MSG,'TID',memory.cur_TID,'write_nlist',memory.multicast_write_nlist,'var_to_be_written',memory.variables_to_write),memory.signal_strength));

    %Initialize read lock array.
    memory.acq_wlock_acks = [];
    for i=1:length(memory.multicast_write_nlist)
        memory.acq_wlock_acks = [memory.acq_wlock_acks inf];
    end    
    
    %Start a timer for the sender to wait for all the ACKS
    Set_Alarm_Receive_Write_Locks_Acks_Timer(t+ (length(memory.multicast_write_nlist)+ 1)*WACK_DELTA);
    
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    for i=1:length(memory.multicast_write_nlist)
        %fprintf(fid, '%d:%d:%s:%d:%c\n', memory.cur_TID,t,'ACQ RLOCK',memory.multicast_read_nlist(i),memory.variables_to_read(i));
    end
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%    

function SendReleaseReadLocks(ID,t,memory,RELEASE_READ_LOCKS_MSG)
    PrintMessage('REL RL');
    disp1(sprintf('%d: Sending release read locks',ID));

    Send_Packet(radiostream(struct('ID',ID, 'type',RELEASE_READ_LOCKS_MSG,'TID',memory.cur_TID,'read_nlist',memory.multicast_read_nlist),memory.signal_strength));

function SendReleaseWriteLock(ID,t,memory,RELEASE_WRITE_LOCK_MSG)
    PrintMessage('REL WL');
    disp1(sprintf('%d: Sending release write locks',ID));

    Send_Packet(radiostream(struct('ID',ID, 'type',RELEASE_WRITE_LOCK_MSG,'TID',memory.cur_TID,'write_nlist',memory.multicast_write_nlist),memory.signal_strength));
    
function disp1(string)
    %disp(string);

function r=rand023()
    ran = rand;
    if ran <= 0.333
        r=0;
    elseif ran >0.333 && ran <=0.667
        r=2;
    else
        r=3;
    end
    
function printIntArray(array)
   disp1(sprintf('%d-',array));
