function application(S)
% ***	
% ***	 Copyright 2006, SUNY Buffalo. All rights reserved.
% ***
% ***    This program is distributed in the hope that it will be useful,
% ***    but WITHOUT ANY WARRANTY; without even the implied warranty of
% ***    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
% ***
% ***

% Written by Muzammil Hussain, mh69@cse.buffalo.edu, Initially adapted from ISIS Prowler
% code
% 


% DO NOT edit simulator code (lines that begin with S;)


S; %%%%%%%%%%%%%%%%%%%   housekeeping  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
S;      persistent app_data 
S;      global ID t
S;      [t, event, ID, data]=get_event(S);
S;      [topology, mote_IDs]=prowler('GetTopologyInfo');
S;      ix=find(mote_IDs==ID);
S;      if ~strcmp(event, 'Init_Application') 
S;         try memory=app_data{ix}; catch memory=[]; end, 
S;      end
S; %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                                          %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%         APPLICATION STARTS               %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%               HERE                       %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                                          %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv%

global fid;
global array_excited_nodes;
global num_finished_nodes;

TRUE=1;
FALSE = 0;

%Message types
READ_REQ_MSG = 1;
READ_ANS_MSG = 2;
WRITE_ALL_REQ_MSG = 3;

MICA2_PACKET_LENGTH_BYTE = 36; % From BMAC paper
sim_params('set','MAC_PACKET_LENGTH',MICA2_PACKET_LENGTH_BYTE * 8);
sim_params('set','RADIO_SS_VAR_CONST',0);
sim_params('set','RADIO_SS_VAR_RAND',0);
sim_params('set','RECEPTION_LIMIT',0.22);   
sim_params('set','TR_ERROR_PROB',0);
sim_params('set','MAC_MIN_WAITING_TIME',200);
sim_params('set','MAC_RAND_WAITING_TIME',128);
sim_params('set','MAC_MIN_BACKOFF_TIME',sim_params('get','MAC_PACKET_LENGTH')*2);
sim_params('set','MAC_RAND_BACKOFF_TIME',sim_params('get','MAC_PACKET_LENGTH')*3);


WT_MAX = sim_params('get','MAC_MIN_WAITING_TIME') + sim_params('get','MAC_RAND_WAITING_TIME');
BT_MAX = sim_params('get','MAC_MIN_BACKOFF_TIME') + sim_params('get','MAC_RAND_BACKOFF_TIME');

NO_OF_RETRANSMISSIONS = 3; %>=1 $$$$  if 1 then no retransmissions

%Spacing between message requests
READ_REQUEST_SPACING = NO_OF_RETRANSMISSIONS*WT_MAX + (NO_OF_RETRANSMISSIONS-1)*BT_MAX + sim_params('get','MAC_PACKET_LENGTH');

MINUTE_IN_BIT_TIME = 40000;
MAX_NUM_NEIGHBOURS = 8;
MAX_NUM_TRANSACTIONS = 1;

var_name_list = 'a';
var_value_list = 1:length(var_name_list);
VAR_RANGE_LENGTH = length(var_name_list);

switch event
case 'Init_Application'
    signal_strength=1;
    
    %%%%%%%%%%%%%%   Memory should be initialized here  %%%%%%%%%%%%%%%%%
    memory=struct('signal_strength', signal_strength,'multicast_read_nlist',[],'variables_to_read',[],'multicast_write_nlist',[],'variables_to_write',[],'values_to_write',[],'answering_nodes_index',[],'write_acks',[],'var_list',[],'cack_var_ref_list',[],'read_req_ref_struct',[],'cancel_acks',[],'cur_TID',-1,'aborting_transaction_flag',-1,'snoop_read_table',[],'snoop_rwi_table',[],'snoop_commit_timer_queue',[],'read_ans_recvd',-1,'tran_count',-1,'num_tran_max',-1,'num_cancel_msg_resend',-1);
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
   array_excited_nodes = sim_params('get_app','ENA');
    
   %array_excited_nodes = [13];

    if ID == 1
        fid = sim_params('get_app','log_file_fid1');
        %array_excited_nodes = [1 2 3 4 5];
        %array_excited_nodes = GetRandomArray(5,25);
    end
    
    if(isempty(array_excited_nodes)) error(['Error ! No nodes are excited!!!']); end
    
    % Initialize all nodes to have variables a to z
    % Assign values from  1 to 24 for each

    memory = SetUpVariables(memory,var_name_list,var_value_list*ID);

    memory.cur_servicing_TID = -1;
    
    if(~isempty(find(array_excited_nodes == ID)))
        memory.num_tran_max = MAX_NUM_TRANSACTIONS;

        memory.tran_count = 1;
        %Initialzing the TID for the new transaction
        memory.cur_TID = str2num(strcat(num2str(ID),num2str(t)));
        disp1(sprintf('%d:%d: Starting new transaction %d with TID : %d',ID,t, memory.tran_count, memory.cur_TID)); 

        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        %fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'START');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 

        grid_size =  sim_params('get','GRID_SQUARE_SIZE_LENGTH');
        memory.neigh_list = GetNeighbours(ID, grid_size);
        memory = DecideReadWriteNeighbours(ID,memory);
        
        %In case it is a read-write thread structure
        if ~isempty(memory.multicast_read_nlist)
            %Send the read request
            memory=SendReadReq(ID,t,memory,READ_REQ_MSG,READ_REQUEST_SPACING,fid);
        else
            %Send the write request
            %Write values of the own same name variables to neighbours
            values_to_write = [];
            for i = 1:length(memory.multicast_write_nlist)
                for j = 1:length(memory.var_list)
                    if memory.variables_to_write(i) == memory.var_list(j).name
                        memory.values_to_write(i) = memory.var_list(j).value;
                    end
                end
            end
            memory=SendWriteReq(ID,t,memory,WRITE_ALL_REQ_MSG,fid);
        end
    end


case 'Packet_Sent'

case 'Packet_Received'
    %LED('red off');
    %LED('green on');
    msg=data.data;
    disp1(sprintf('%d:%d: Received packet of type %d',ID,t,msg.type));
    switch msg.type
        case READ_REQ_MSG
            % Check if the receiving node should service a read request
            if ~(isempty(find(msg.read_nlist == ID)))
                index = find(msg.read_nlist == ID);

                new_i = length(memory.read_req_ref_struct) + 1;
                memory.read_req_ref_struct(new_i).requesting_node = msg.ID;
                memory.read_req_ref_struct(new_i).requesting_tid = msg.TID;
                memory.read_req_ref_struct(new_i).requested_variable = msg.var_to_be_read(index);
                memory.read_req_ref_struct(new_i).read_ans_exp_time = t + READ_REQUEST_SPACING*(index-1);
                
                Set_Alarm_Send_Read_Request_Timer(t + READ_REQUEST_SPACING*(index-1));
            end
                        
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case READ_REQ_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        case READ_ANS_MSG
            prev_read_answers_not_recvd = FALSE;
            
            if msg.dest_id == ID && ~(isempty(find(memory.multicast_read_nlist == msg.ID))) %&& msg.TID == memory.cur_TID
                index = find(memory.multicast_read_nlist == msg.ID);
                %disp1(sprintf('%d:%d: Received answer whose Index in the read request : %d',ID,t,index));
                % Check if it has come in the right order as specified in
                % the read request
                for i = 1:index-1    
                    if memory.answering_nodes_index(i).flag == inf
                        prev_read_answers_not_recvd = TRUE;
                    end
                end

                if memory.answering_nodes_index(index).flag ~= inf
                    error(['%d trying to overwrite' msg.ID])
                end

                % Record the answer
                disp1(sprintf('%d:%d Recording answer from node %d',ID,t,msg.ID));
                memory.answering_nodes_index(index).name = msg.variable_names;
                memory.answering_nodes_index(index).value = msg.variable_values;
                memory.answering_nodes_index(index).flag = 1;

                %In case all requested nodes have replied send the
                %writeall
                if index == length(memory.multicast_read_nlist) && (prev_read_answers_not_recvd == FALSE)
                    memory.read_ans_recvd = 1;
                    PrintMessage('WR ALL');
                    
                    result=ApplyTranformation(memory.multicast_write_nlist,'MAX');
                    %Set each variable in the neighbor list to be
                    %written the same value
                    for i=1:length(memory.multicast_write_nlist)
                        memory.values_to_write(i) = result;
                    end

                    %Send the write req
                    memory=SendWriteReq(ID,t,memory,WRITE_ALL_REQ_MSG,fid);
                end
            end

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case READ_ANS_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        case WRITE_ALL_REQ_MSG
            if ~(isempty(find(msg.write_nlist == ID))) 
                index = find(msg.write_nlist == ID);

                for i=1:length(memory.var_list)
                    if memory.var_list(i).name == msg.variables_to_write(index)
                        memory.var_list(i).new_value = msg.values_to_write(index);
                        
                        memory.var_list(i).updating_tid = [memory.var_list(i).updating_tid msg.TID];
                        memory.var_list(i).updating_node = [memory.var_list(i).updating_node msg.ID];
                        
                        disp1(sprintf('%d:%d: Updating variable %c to new value %d supplied by %d for transaction %d',ID,t,memory.var_list(i).name,memory.var_list(i).new_value(end),msg.ID,msg.TID));
                        break;
                    end
                end
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%case WRITE_ALL_REQ_MSG%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        otherwise
            error(['Bad msg type received: ' msg.type])
    end
case 'Collided_Packet_Received'
    disp1(sprintf('%d:%d: Received collided packet !!',ID,t));
    %LED('red on');
    %LED('green off');
    %LED('yellow off');
    
case 'Send_Read_Answer_Timer_Expired'
    % Send the values of the variables as ordered in memory.read_requested_variables
    disp1(sprintf('%d:%d: Read Answer Timer Expired !!!',ID,t));
    
    index = -1;
    for i = 1:length(memory.read_req_ref_struct)
        if memory.read_req_ref_struct(i).read_ans_exp_time == t
            index = i;
            break;
        end
    end
    
    found = FALSE;
    for i=1:length(memory.var_list)
        if memory.var_list(i).name == memory.read_req_ref_struct(index).requested_variable
            disp1(sprintf('%d:%d: Reading variable %c which gives value %d',ID,t,memory.var_list(i).name,memory.var_list(i).value));
            PrintMessage('R ANS');
            Send_Packet(radiostream(struct('ID',ID, 'type',READ_ANS_MSG,'dest_id',memory.read_req_ref_struct(index).requesting_node,'TID',memory.read_req_ref_struct(index).requesting_tid,'variable_names',memory.var_list(i).name,'variable_values',memory.var_list(i).value),memory.signal_strength));
            disp1(sprintf('%d:%d: Sending answer to %d for transaction %d',ID,t,memory.read_req_ref_struct(index).requesting_node,memory.read_req_ref_struct(index).requesting_tid));
            found = TRUE;
            %Remove the head
            memory.read_req_ref_struct(index) = [];
            break;
        end
    end
        
    if found == FALSE
        error(['Error ! Read request variable does not exist in node'])
    end

    
case 'Receive_Read_Answers_Timer_Expired'
    if memory.read_ans_recvd == 0
        disp1(sprintf('%d:%d: Read Request timing out !!!',ID,t));
        disp1(sprintf('%d:%d: Canceling transaction %d !!!',ID,t,memory.cur_TID));
        
        PrintMessage('CNCL R');

        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        %fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'TIMEOUT-CANCEL');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%        

        memory.cur_TID = str2num(strcat(num2str(ID),num2str(t)));
        disp1(sprintf('%d:%d: Starting %d new transaction with TID : %d',ID,t, memory.tran_count, memory.cur_TID));
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
        %fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'START');
        %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 

        memory=SendReadReq(ID,t,memory,READ_REQ_MSG,READ_REQUEST_SPACING,fid);
    end

        
case 'GuiInfoRequest'
    if ~isempty(memory)
        disp1(sprintf('Memory Dump of mote ID# %d:\n',ID)); disp1(memory)
    else
        disp1(sprintf('No memory dump available for node %d.\n',ID)); 
    end
    
case 'Application_Stopped'
   % this event is called when simulation is stopped/suspended 
   
case 'Application_Finished'
   % this event is called when simulation is finished 
    
otherwise
    error(['Bad event name for application: ' event])
end

%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                                          %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%         APPLICATION ENDS                 %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%               HERE                       %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                                          %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
S; %%%%%%%%%%%%%%%%%%%%%% housekeeping %%%%%%%%%%%%%%%%%%%%%%%%%%%
S;        app_data{ix}=memory;
S; %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                           %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%        COMMANDS           %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%                           %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function b=Send_Packet(data);
global ID t
radio=prowler('GetRadioName');
b=feval(radio, 'Send_Packet', ID, data, t);

function b=Set_Alarm_Send_Read_Request_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Send_Read_Answer_Timer_Expired', ID));

function b=Set_Alarm_Receive_Read_Answers_Timer(alarm_time);
global ID
prowler('InsertEvents2Q', make_event(alarm_time, 'Receive_Read_Answers_Timer_Expired', ID));

function PrintMessage(msg)
global ID
prowler('TextMessage', ID, msg)

function LED(msg)
global ID
prowler('LED', ID, msg)

function DrawLine(command, varargin)
switch lower(command)
case 'line'
    prowler('DrawLine', varargin{:})
case 'arrow'
    prowler('DrawArrow', varargin{:})
case 'delete'
    prowler('DrawDelete', varargin{:})
otherwise
    error('Bad command for DrawLine.')
end

function array =GetRandomArray(n,max)
    count = 1;
    array = [];
    while(count<n+1)
      x = ceil(max*rand);
      if isempty(find(array == x))
        array(count) = x;
        count = count + 1;
      end
    end
    
%x is the number of randomly chosen nodes needed
%N is the side of the N*N grid
%For example (5,5) would return an array of 5 randomly chosen nodes from a
%square grid of 25 nodes
function array = GetNeighbours(x,N)
    array = [];
    left = x - N;
    right = x + N;
    % Add the left neighbour if possible
    if(left >= 1)
        array = [array left];
    end
    % Add the right neighbour if possible
    if(right <= N*N)
        array = [array right];
    end
    
    %Add the northern/sourthern neighbour
    if( mod(x, N) == 0)
        array = [array x-1]; % Add southern neighbour if x is topmost
    elseif(mod(x, N) == 1)
        array = [array x+1];
    else
        array = [array x-1 x+1];
    end
    
    if(left >= 1)
        %Add the northern/sourthern neighbour of left
        if( mod(left, N) == 0)
            array = [array left-1]; % Add southern neighbour if x is topmost
        elseif(mod(left,N) == 1)
            array = [array left+1];
        else
            array = [array left-1 left+1];
        end
    end 
    
    if(right <= N*N)
        %Add the northern/sourthern neighbour of right
        if( mod(right,N) == 0)
            array = [array right-1]; % Add southern neighbour if x is topmost
        elseif(mod(right,N) == 1)
            array = [array right+1];
        else
            array = [array right-1 right+1];
        end
    end     
    array = sort(array);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function mem=SetUpVariables(mem,name_list,value_list)
    len = length(name_list);
    for i = 1 : len
        mem.var_list(i).name = name_list(i);
        mem.var_list(i).value = value_list(i);
        mem.var_list(i).new_value = [];
        mem.var_list(i).commit_time = [];
        mem.var_list(i).w_ack_send_time = [];
        mem.var_list(i).updating_tid = [];
        mem.var_list(i).updating_node = [];
        
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function memory = DecideReadWriteNeighbours(ID,memory)

    %Decide the read neighbours
    memory = DecideReadNeighbours(ID,memory);
    while(length(memory.multicast_read_nlist) == 0)
        memory = DecideReadNeighbours(ID,memory);
    end
    
    %Decide the write neighbours
    memory = DecideWriteNeighbours(ID,memory);
    while(length(memory.multicast_write_nlist) == 0)
        memory = DecideWriteNeighbours(ID,memory);
    end
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function memory = DecideReadNeighbours(ID,memory)
    read_trans_matrix = zeros(1,length(memory.neigh_list));
    
    %Decide the read neighbours
    for i=1:length(read_trans_matrix)
        %Randomly choose the read neigbours with 1 being they are in the
        %list
        read_trans_matrix(i) = round(rand);
        if (read_trans_matrix(i) == 1)
            memory.variables_to_read = [memory.variables_to_read 'a'];
        end
    end
    memory.multicast_read_nlist = memory.neigh_list(find(read_trans_matrix == 1));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function memory = DecideWriteNeighbours(ID,memory)
    write_trans_matrix = zeros(1,length(memory.multicast_read_nlist));
    
    %Decide the write neighbours
    for i=1:length(write_trans_matrix)
        %If there is only one read neighbour then write to that without
        %choosing
        if length(memory.multicast_read_nlist) == 1
            write_trans_matrix(1) = 1;
            memory.variables_to_write = [memory.variables_to_write 'a'];
            break;
        else
            %Randomly choose the write neigbours with 1 being they are in the
            %list
            write_trans_matrix(i) = round(rand);        
            if (write_trans_matrix(i) == 1)
                memory.variables_to_write = [memory.variables_to_write 'a'];
            end
        end
    end
    memory.multicast_write_nlist = memory.multicast_read_nlist(find(write_trans_matrix == 1));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function result=ApplyTranformation(list,oper_type)
    switch oper_type
        case 'PRODUCT'
            result = prod(list);
        case 'SUM'
            result = sum(list);
        case 'MAX'
            result = max(list);
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
function memory = SendReadReq(ID,t,memory,READ_REQ_MSG,READ_REQUEST_SPACING,fid)
    disp1(sprintf('%d: Sending read request',ID));
    PrintMessage('R REQ');
    Send_Packet(radiostream(struct('ID',ID, 'type',READ_REQ_MSG,'TID',memory.cur_TID,'read_nlist',memory.multicast_read_nlist,'var_to_be_read',memory.variables_to_read),memory.signal_strength));

    memory.read_ans_recvd = 0;
    %Start a timer for the sender to wait for all the ACKS
    Set_Alarm_Receive_Read_Answers_Timer(t+ (length(memory.multicast_read_nlist)+ 1)*READ_REQUEST_SPACING);

    %Initialize the answer array
    memory.answering_nodes_index = [];
    for i=1:length(memory.multicast_read_nlist)
        memory.answering_nodes_index(i).flag = inf;
        memory.answering_nodes_index(i).name = '';
        memory.answering_nodes_index(i).value = inf;
    end
    
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    for i=1:length(memory.multicast_read_nlist)
        %fprintf(fid, '%d:%d:%s:%d:%c\n', memory.cur_TID,t,'READ',memory.multicast_read_nlist(i),memory.variables_to_read(i));
    end
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
function memory = SendWriteReq(ID,t,memory,WRITE_ALL_REQ_MSG,fid)
    PrintMessage('W ALL');
    disp1(sprintf('%d: Sending write req',ID));
    
    %Initialize the write ack receive array
    memory.write_acks = [];
    for i=1:length(memory.multicast_write_nlist)
        memory.write_acks = [memory.write_acks inf];
    end
  
   %Sending the write-all message to subset of neighbors
   Send_Packet(radiostream(struct('ID',ID, 'type',WRITE_ALL_REQ_MSG,'TID',memory.cur_TID,'write_nlist',memory.multicast_write_nlist,'variables_to_write',memory.variables_to_write,'values_to_write',memory.values_to_write),memory.signal_strength));

   %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    write_string = [];
    for i=1:length(memory.multicast_write_nlist)
        write_string = strcat(write_string,sprintf('::%d:%c',memory.multicast_write_nlist(i),memory.variables_to_write(i)));
    end
    %fprintf(fid, '%d:%d:%s:%s\n', memory.cur_TID,t,'WRITE',write_string);
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%

    PrintMessage('END'); 
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$%
    %fprintf(fid, '%d:%d:%s\n', memory.cur_TID,t,'END');
    %$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$% 
    num_finished_nodes = sim_params('get_app','num_finished_nodes');
    num_finished_nodes = num_finished_nodes + 1;
    sim_params('set_app','num_finished_nodes',num_finished_nodes);    
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   

function disp1(string)
%disp(string);
